from typing import Type, Any, TYPE_CHECKING, Union, Optional, Tuple

from pydantic import BaseModel

from .types import ObjectIdStr
from .exceptions import MotordanticValidationError


__all__ = ('validate_field_value', 'sort_validation')


if TYPE_CHECKING:
    from .models import MongoModel


def validate_field_value(cls: Type['MongoModel'], field_name: str, value: Any) -> Any:
    """extra helper value validation

    Args:
        cls (Type[): mongo model class
        field_name (str): name of field
        value (Any): value

    Raises:
        AttributeError: if not field in __fields__
        MongoValidationError: if invalid value type

    Returns:
        Any: value
    """
    if field_name == '_id':
        field = ObjectIdStr()  # type: ignore
    else:
        field = cls.__fields__.get(field_name)  # type: ignore
    error_ = None
    if isinstance(field, ObjectIdStr):
        try:
            value = field.validate(value)
        except ValueError as e:
            error_ = e
    elif not field:
        raise AttributeError(f'invalid field - {field_name}')
    else:
        value, error_ = field.validate(value, {}, loc=field.alias, cls=cls)
    if error_:
        raise MotordanticValidationError([error_], type(value))
    if field_name in cls.__db_refs__:  # type: ignore
        if isinstance(value, list):
            s = [v.to_ref() for v in value]
            return s
        return value.to_ref() if value else None
    else:
        return value.dict() if isinstance(value, BaseModel) else value


def sort_validation(
    sort: Optional[int] = None, sort_fields: Union[list, tuple, None] = None
) -> Tuple[Any, ...]:
    if sort is not None:
        if sort not in (1, -1):
            raise ValueError(f'invalid sort value must be 1 or -1 not {sort}')
        if not sort_fields:
            sort_fields = ('_id',)
    return sort, sort_fields



import pytest
import pytest_asyncio

from bson import ObjectId

from motordantic.models import MongoModel
from motordantic.exceptions import DoesNotExist


class Ticket(MongoModel):
    name: str
    position: int
    config: dict
    sign: int = 1
    type_: str = 'ga'
    array: list = [1, 2]

    class Config:
        excluded_query_fields = ('sign', 'type')

    @property
    def position_property(self):
        return self.position


class Trash(MongoModel):
    name: str
    date: str


@pytest_asyncio.fixture(scope='session', autouse=True)
async def drop_ticket_collection(event_loop):
    yield
    await Ticket.Q.drop_collection(force=True)
    await Trash.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_insert_one():
    data = {
        'name': 'second',
        'position': 2,
        'config': {'param1': 'value2'},
        'array': ['test', 'adv', 'comagic'],
    }
    object_id = await Ticket.Q.insert_one(**data)
    assert isinstance(object_id, ObjectId)


@pytest.mark.asyncio
async def test_insert_many():
    data = [
        {
            'name': 'first',
            'position': 1,
            'config': {'param1': 'value'},
            'array': ['test', 'adv', 'calltouch'],
        },
        {
            'name': 'third',
            'position': 3,
            'config': {'param1': 'value3'},
            'array': ['test', 'adv', 'comagic', 'cost'],
        },
        {
            'name': 'third',
            'position': 4,
            'config': {'param1': 'value3'},
            'array': ['test', 'adv', 'comagic', 'cost', 'trash'],
        },
    ]
    inserted = await Ticket.Q.insert_many(data)
    assert inserted == 3


@pytest.mark.asyncio
async def test_distinct():
    data = await Ticket.Q.distinct('position', name='second')
    assert data == [2]


@pytest.mark.asyncio
async def test_queryset_serialize():
    result = await Ticket.Q.find(name='second')
    data = result.serialize(fields=['name', 'config'])
    assert len(data[0]) == 2
    assert data[0]['config'] == {'param1': 'value2'}
    assert data[0]['name'] == 'second'
    assert isinstance(data, list)


@pytest.mark.asyncio
async def test_find_one():
    await test_insert_one()
    data = await Ticket.Q.find_one(name='second')
    second = await Ticket.Q.find_one(_id=data._id)
    assert isinstance(data, MongoModel)
    assert data.name == 'second'  # type: ignore
    assert data.position == 2  # type: ignore
    assert second._id == data._id
    assert second.position_property == 2
    assert second.data['position_property'] == 2


@pytest.mark.asyncio
async def test_find():
    result = await Ticket.Q.find(limit_rows=1, name='second')
    assert len(result.data) == 1


@pytest.mark.asyncio
async def test_async_get():
    # await .test_insert_one()
    data = await Ticket.Q.get(name='second')
    second = await Ticket.Q.get(_id=data._id)
    assert isinstance(data, MongoModel)
    assert data.name == 'second'  # type: ignore
    assert data.position == 2  # type: ignore
    assert second._id == data._id
    with pytest.raises(DoesNotExist):
        _ = await Ticket.Q.get(name='invalid_name')


@pytest.mark.asyncio
async def test_save():
    obj = await Ticket.Q.find_one(name='second')
    obj.position = 2310
    obj.name = 'updated'
    await obj.save()
    none_obj = await Ticket.Q.find_one(name='second', position=222222)
    assert none_obj is None
    new_obj = await Ticket.Q.find_one(_id=obj._id)
    assert new_obj.name == 'updated'
    assert new_obj.position == 2310

    obj.name = 'second'
    obj.position = 2

    await obj.save()
    last_obj = await Ticket.Q.find_one(name='second')
    assert last_obj.name == 'second'
    assert last_obj.position == 2


@pytest.mark.asyncio
async def test_update_one():
    updated = await Ticket.Q.update_one(name='second', config__set={'updated': 1})
    data = await Ticket.Q.find_one(name='second')
    assert updated == 1
    assert data.config == {'updated': 1}


@pytest.mark.asyncio
async def test_update_many():
    data = await Ticket.Q.update_many(name='third', config__set={'updated': 3})
    updated = await Ticket.Q.find_one(name='third')
    assert data == 2
    assert updated.config == {'updated': 3}


@pytest.mark.asyncio
async def test_gte_lte_in_one_field():

    await Trash.Q.insert_one(name='first', date='2022-01-01')
    await Trash.Q.insert_one(name='second', date='2022-01-01')
    await Trash.Q.insert_one(name='third', date='2022-01-03')
    await Trash.Q.insert_one(name='4', date='2022-01-05')

    data = await Trash.Q.count(
        date__gte='2022-01-01',
        date__lte='2022-01-03',
    )
    assert data == 3


@pytest.mark.asyncio
async def test_find_with_regex():
    second = await Ticket.Q.find_one(name__regex='ond')
    assert second.name == 'second'

    starts = await Ticket.Q.find_one(name__startswith='s')
    assert starts._id == second._id

    istarts = await Ticket.Q.find_one(name__istartswith='S')
    assert istarts._id == second._id

    not_istarts = await Ticket.Q.find_one(name__not_istartswith='S', position=1)
    assert not_istarts.name == 'first'


@pytest.mark.asyncio
async def test_find_and_update():
    data_default = await Ticket.Q.find_one_and_update(name='second', position__set=23)
    assert data_default.position == 23

    data_with_prejection = await Ticket.Q.find_one_and_update(
        name='first', position__set=12, projection_fields=['position']
    )
    assert isinstance(data_with_prejection, dict)
    assert data_with_prejection['position'] == 12


@pytest.mark.asyncio
async def test_delete_one():
    deleted = await Ticket.Q.delete_one(name='second')
    assert deleted == 1


@pytest.mark.asyncio
async def test_delete_many():
    deleted = await Ticket.Q.delete_many(name='second')
    assert deleted == 1



from typing import Optional, Any, Callable, cast

from pydantic.fields import (
    Field as PydanticField,
    FieldInfo as PydanticFieldInfo,
    Undefined,
)

__all__ = ('Field', 'FieldInfo')


class FieldInfo:
    __slots__ = ("pydantic_field_info", "db_field_name", "default")

    def __init__(
        self,
        *,
        pydantic_field_info: PydanticFieldInfo,
        db_field_name: Optional[str],
    ):
        self.pydantic_field_info = pydantic_field_info
        self.db_field_name = db_field_name


def Field(
    default: Any = Undefined,
    *,
    db_field_name: Optional[str] = None,
    default_factory: Optional[Callable[[], Any]] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    **extra: Any,
) -> Any:
    pydantic_field = PydanticField(
        default,
        default_factory=default_factory,
        # alias=db_field_name,
        title=cast(str, title),
        description=cast(str, description),
        const=cast(bool, const),
        gt=cast(float, gt),
        ge=cast(float, ge),
        lt=cast(float, lt),
        le=cast(float, le),
        multiple_of=cast(float, multiple_of),
        min_items=cast(int, min_items),
        max_items=cast(int, max_items),
        min_length=cast(int, min_length),
        max_length=cast(int, max_length),
        regex=cast(str, regex),
        **extra,
    )
    return FieldInfo(pydantic_field_info=pydantic_field, db_field_name=db_field_name)



from typing import TYPE_CHECKING

from .exceptions import MotordanticValidationError

if TYPE_CHECKING:
    from .models import MongoModel


__all__ = ('Sum', 'Avg', 'Min', 'Count', 'Max')


class BaseAggregation(object):
    """Abstract class for Aggregation"""

    _operation: str

    def __init__(self, field: str):
        self.field = field

    @property
    def operation(self) -> str:
        if not self._operation:
            raise NotImplementedError('implement _operation')
        return self._operation

    def _validate_field(self, mongo_model: 'MongoModel'):
        if self.field not in mongo_model.__fields__ and self.field != '_id':
            raise MotordanticValidationError(
                f'invalid field "{self.field}" for this model, field must be one of {list(mongo_model.__fields__.keys())}'
            )

    def _aggregate_query(self, mongo_model: 'MongoModel') -> dict:
        self._validate_field(mongo_model)
        query = {
            f'{self.field}__{self.operation}': {f'${self.operation}': f'${self.field}'}
        }
        return query


class Sum(BaseAggregation):
    """
    Simple sum aggregation

    generated query: {'field__sum': {'$sum': 'field'}}

    return: {'field__sum': <value>}
    """

    _operation: str = 'sum'


class Max(BaseAggregation):
    """
    Simple max aggregation

    generated query: {'field__max': {'$max': 'field'}}

    return: {'field__max': <value>}
    """

    _operation: str = 'max'


class Min(BaseAggregation):
    """
    Simple min aggregation

    generated query: {'field__min': {'$min': 'field'}}

    return: {'field__min': <value>}
    """

    _operation: str = 'min'


class Avg(BaseAggregation):
    """
    Simple avg aggregation

    generated query: {'field__avg': {'$avg': 'field'}}

    return: {'field__avg': <value>}
    """

    _operation: str = 'avg'


class Count(BaseAggregation):
    """
    Simple Count aggregation

    generated query:
        - if field = _id
            {
                '_id': None
                'count': {'$sum': 1},
            }
        - else
            {
                '_id': 'field'
                'count': {'$sum': 1},
            }
    return: {'_id': <value>, 'count': value}
    """

    _operation: str = 'count'

    def _aggregate_query(self, mongo_model: 'MongoModel') -> dict:
        self._validate_field(mongo_model)
        query = {
            "_id": f'${self.field}' if self.field != '_id' else None,
            f'count': {'$sum': 1},
        }
        return query



import json
from typing import (
    Dict,
    Any,
    Union,
    Optional,
    List,
    Tuple,
    Set,
    TYPE_CHECKING,
    Type,
    ClassVar,
)

from bson import ObjectId, DBRef
from motor.core import AgnosticClientSession, AgnosticCollection, AgnosticDatabase

from pydantic.main import ModelMetaclass as PydanticModelMetaclass
from pydantic import BaseModel as BasePydanticModel, root_validator
from pymongo import IndexModel


from .relation import RelationManager
from .types import ObjectIdStr, RelationInfo, Relation
from .exceptions import (
    NotDeclaredField,
    MotordanticInvalidArgsParams,
    MotordanticValidationError,
    MotordanticConnectionError,
)
from .property import classproperty
from .validaton import validate_field_value
from .extra import ExtraQueryMapper, take_relation
from .query import Query, QueryCombination
from .querybuilder import QueryBuilder

__all__ = ('MongoModel', 'ModelMetaclass')

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny
    from pydantic.typing import AbstractSetIntStr  # noqa: F401


_is_mongo_model_class_defined = False


class ModelMetaclass(PydanticModelMetaclass):
    def __new__(mcs, name, bases, namespace, **kwargs):  # type: ignore
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        indexes = set()
        if _is_mongo_model_class_defined and issubclass(cls, MongoModel):
            querybuilder = getattr(cls, '__querybuilder__')
            if querybuilder is None:
                querybuilder = QueryBuilder(cls)  # type: ignore
                setattr(cls, '__querybuilder__', querybuilder)
            db_refs = {}
            for k, v in cls.__fields__.items():
                relation_info = take_relation(v)
                if relation_info is not None:
                    db_refs[k] = relation_info
            setattr(cls, '__db_refs__', db_refs)
            if db_refs:
                setattr(cls, '__relation_manager__', RelationManager(cls))
        json_encoders = getattr(cls.Config, 'json_encoders', {})  # type: ignore
        json_encoders.update({ObjectId: lambda f: str(f)})
        setattr(cls.Config, 'json_encoders', json_encoders)  # type: ignore
        exclude_fields = getattr(cls.Config, 'exclude_fields', tuple())  # type: ignore
        setattr(cls, '__indexes__', indexes)
        setattr(cls, '__database_exclude_fields__', exclude_fields)
        return cls


class MongoModel(BasePydanticModel, metaclass=ModelMetaclass):
    __database__: Optional[AgnosticDatabase] = None
    __collection__: Optional[AgnosticCollection] = None
    __indexes__: Set['str'] = set()
    __database_exclude_fields__: Union[Tuple, List] = tuple()
    __querybuilder__: Optional[QueryBuilder] = None
    __db_refs__: ClassVar[Optional[Dict[str, RelationInfo]]] = None
    __relation_manager__: Optional[RelationManager] = None
    _id: Optional[ObjectIdStr] = None

    def __setattr__(self, key, value):
        if key == '_id':
            self.__dict__[key] = value
            return value
        else:
            return super().__setattr__(key, value)

    @classproperty
    def relation_manager(cls) -> Optional[RelationManager]:
        return cls.__relation_manager__

    @classmethod
    async def ensure_indexes(cls):
        """method for create/update/delete indexes if indexes declared in Config property"""

        indexes = getattr(cls.__config__, 'indexes', [])
        if not all([isinstance(index, IndexModel) for index in indexes]):
            raise ValueError('indexes must be list of IndexModel instances')
        if indexes:
            db_indexes = await cls.Q.list_indexes()
            indexes_to_create = [
                i for i in indexes if i.document['name'] not in db_indexes
            ]
            indexes_to_delete = [
                i
                for i in db_indexes
                if i not in [i.document['name'] for i in indexes] and i != '_id_'
            ]
            result = []
            if indexes_to_create:
                try:
                    result = await cls.Q.create_indexes(indexes_to_create)
                except MotordanticConnectionError:
                    pass
            if indexes_to_delete:
                for index_name in indexes_to_delete:
                    await cls.Q.drop_index(index_name)
                db_indexes = await cls.Q.list_indexes()
            indexes = set(list(db_indexes.keys()) + result)
        setattr(cls, '__indexes__', indexes)

    @classmethod
    def _get_properties(cls) -> list:
        return [
            prop
            for prop in dir(cls)
            if prop
            not in (
                "__values__",
                "fields",
                "data",
                "_connection",
                "_collection_name",
                "_collection",
                "querybuilder",
                "Q",
                "AQ",
                "pk",
                "_query_data",
                "fields_all",
                "all_fields",
            )
            and isinstance(getattr(cls, prop), property)
        ]

    @classmethod
    def parse_obj(cls, data: Any) -> Any:
        obj = super().parse_obj(data)
        if '_id' in data:
            obj._id = data['_id'].__str__()
        return obj

    async def save(
        self,
        updated_fields: Union[Tuple, List] = [],
        session: Optional[AgnosticClientSession] = None,
    ) -> Any:
        if self._id is not None:
            data = {
                '_id': (
                    self._id if isinstance(self._id, ObjectId) else ObjectId(self._id)
                )
            }
            if updated_fields:
                if not all(field in self.__fields__ for field in updated_fields):
                    raise MotordanticValidationError('invalid field in updated_fields')
            else:
                updated_fields = tuple(self.__fields__.keys())
            for field in updated_fields:
                data[f'{field}__set'] = getattr(self, field)
            await self.Q.update_one(
                session=session,
                **data,
            )
            return self
        data = {
            field: value
            for field, value in self.__dict__.items()
            if field in self.__fields__
        }
        object_id = await self.Q.insert_one(
            session=session,
            **data,
        )
        self._id = object_id
        return self

    @classmethod
    def __validate_field(cls, field: str) -> bool:
        if field not in cls.__fields__ and field != '_id':
            raise NotDeclaredField(field, list(cls.__fields__.keys()))
        elif field in cls.__database_exclude_fields__:
            return False
        return True

    @classmethod
    def _check_query_args(
        cls,
        logical_query: Union[
            List[Any], Dict[Any, Any], str, Query, QueryCombination, None
        ] = None,
    ) -> 'DictStrAny':
        """check if query = Query obj or QueryCombination

        Args:
            logical_query (Union[ List[Any], Dict[Any, Any], str, Query, QueryCombination ], optional): Query | QueryCombination. Defaults to None.

        Raises:
            InvalidArgsParams: if not Query | QueryCombination

        Returns:
            Dict: generated query dict
        """
        if not isinstance(logical_query, (QueryCombination, Query)):
            raise MotordanticInvalidArgsParams()
        return logical_query.to_query(cls)  # type: ignore

    @classmethod
    def use(cls: Type["MongoModel"], database: AgnosticDatabase) -> None:
        assert database is not None
        cls.__database__ = database

    @classmethod
    def _parse_extra_params(cls, extra_params: List) -> tuple:
        field_param, extra = [], []
        methods = ExtraQueryMapper.methods
        for param in extra_params:
            if param in methods:
                extra.append(param)
            else:
                field_param.append(param)
        return field_param, extra

    @classmethod
    def _validate_query_data(cls, query: Dict) -> 'DictStrAny':
        """main validation method

        Args:
            query (Dict): basic query

        Returns:
            Dict: parsed query
        """
        data: dict = {}
        for query_field, value in query.items():
            field, *extra_params = query_field.split("__")
            inners, extra_params = cls._parse_extra_params(extra_params)
            if not cls.__validate_field(field):
                continue
            extra = ExtraQueryMapper(cls, field).query(extra_params, value)
            if extra:
                value = extra[field]
            elif field == '_id':
                value = ObjectId(value)
            else:
                value = validate_field_value(cls, field, value) if not inners else value
            if inners:
                field = f'{field}.{".".join(i for i in inners)}'
            if (
                extra
                and field in data
                and ('__gt' in query_field or '__lt' in query_field)
            ):
                data[field].update(value)
            else:
                data[field] = value
        return data

    @classproperty
    def database(cls) -> AgnosticDatabase:
        """Returns the database that is currently associated with this document."""
        if not hasattr(cls, "__database__") or cls.__database__ is None:
            raise AttributeError("Accessing database without using it first.")
        return cls.__database__

    @classproperty
    def collection(cls: 'MongoModel') -> AgnosticCollection:
        """Returns the collection for this :class:`MongoModel`."""
        if (
            cls.__collection__ is None
            or cls.__collection__.database is not cls.database
        ):
            cls.__collection__ = cls.database.get_collection(cls.set_collection_name())  # type: ignore
        return cls.__collection__

    @classproperty
    def fields_all(cls) -> list:
        """return all fields with properties(not document fields)"""
        fields = list(cls.__fields__.keys())
        return_fields = fields + cls._get_properties()
        return return_fields

    @classproperty
    def Q(cls) -> Optional[QueryBuilder]:
        return cls.__querybuilder__

    def dict(  # type: ignore
        self,
        *,
        include: Optional['AbstractSetIntStr'] = None,
        exclude: Optional['AbstractSetIntStr'] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        with_props: bool = True,
        to_query: bool = True,
    ) -> 'DictStrAny':
        """
        Generate a dictionary representation of the model, optionally specifying which fields to include or exclude.

        """
        attribs = super().dict(
            include=include,  # type: ignore
            exclude=exclude,  # type: ignore
            by_alias=by_alias,
            skip_defaults=skip_defaults,  # type: ignore
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        if with_props:
            props = self._get_properties()
            # Include and exclude properties
            if include:
                props = [prop for prop in props if prop in include]
            if exclude:
                props = [prop for prop in props if prop not in exclude]

            # Update the attribute dict with the properties
            if props:
                attribs.update({prop: getattr(self, prop) for prop in props})
        if self.has_db_refs():
            for field in self.__db_refs__:  # type: ignore
                attrib_data = attribs[field]
                if attrib_data and not isinstance(attrib_data, dict):
                    attribs[field] = (
                        attrib_data.to_dict()
                        if not isinstance(attrib_data, list)
                        else [
                            a.to_dict() if not isinstance(a, dict) else a
                            for a in attrib_data
                        ]
                    )
        return attribs

    @classmethod
    def to_relation(cls, object_id: Union[str, ObjectId]) -> Relation:
        if isinstance(object_id, str):
            object_id = ObjectId(object_id)
        db_ref = DBRef(collection=cls.set_collection_name(), id=object_id)
        return Relation(db_ref, cls)

    @classmethod
    def has_db_refs(cls) -> bool:
        return bool(cls.__db_refs__)

    @property
    def data(self) -> 'DictStrAny':
        return self.dict(with_props=True)

    @property
    def _query_data(self) -> 'DictStrAny':
        return self.dict(with_props=False)

    @classmethod
    def set_collection_name(cls) -> str:
        """main method for set collection

        Returns:
            str: collection name
        """
        return cls.__name__.lower()

    def serialize(self, fields: Union[Tuple, List]) -> 'DictStrAny':
        data: dict = self.dict(include=set(fields))
        return {f: data[f] for f in fields}

    def serialize_json(self, fields: Union[Tuple, List]) -> str:
        return json.dumps(self.serialize(fields))

    @property
    def pk(self):
        return self._id

    @root_validator
    def validate_all_fields(cls, values):
        for field, value in values.items():
            if isinstance(value, MongoModel):
                raise ValueError(
                    f'{field} - cant be instance of MongoModel without Relation'
                )
        return values


_is_mongo_model_class_defined = True



from time import sleep
from types import GeneratorType
from typing import Callable, Any

from pymongo.errors import (
    ServerSelectionTimeoutError,
    AutoReconnect,
    NetworkTimeout,
    ConnectionFailure,
)


class BaseMotorDanticException(Exception):
    pass


class NotDeclaredField(BaseMotorDanticException):
    def __init__(self, field_name: str, fields: list, *args):
        self.field_name = field_name
        self.fields = fields
        super().__init__(*args)

    def __str__(self):
        return f"This field - {self.field_name} not declared in {self.fields}"


class DoesNotExist(BaseMotorDanticException):
    def __init__(self, model_name: str, *args):
        super().__init__(args)
        self.model_name = model_name

    def __str__(self):
        return f'row does not exist for model: {self.model_name}'


class MotordanticValidationError(BaseMotorDanticException):
    pass


class MotordanticInvalidArgsParams(BaseMotorDanticException):
    def __str__(self):
        return 'Arguments must be Query objects'


class MotordanticConnectionError(BaseMotorDanticException):
    pass


class MotordanticIndexError(BaseMotorDanticException):
    pass


def handle_and_convert_connection_errors(func: Callable) -> Any:
    """decorator for handle connection errors and raise MongoConnectionError

    Args:
        func (Callable):any query to mongo

    Returns:
        Any: data
    """

    def generator_wrapper(generator):
        yield from generator

    def main_wrapper(*args, **kwargs):
        counter = 1
        while True:
            try:
                result = func(*args, **kwargs)
                if isinstance(result, GeneratorType):
                    result = generator_wrapper(result)
                return result
            except (
                AutoReconnect,
                ServerSelectionTimeoutError,
                NetworkTimeout,
                ConnectionFailure,
            ) as e:
                counter += 1
                if counter > 3:
                    raise MotordanticConnectionError(str(e))
                sleep(counter)

    return main_wrapper



import pytest
import pytest_asyncio

from motordantic.models import MongoModel
from motordantic.query import Query


class TicketForQuery(MongoModel):
    name: str
    position: int


@pytest_asyncio.fixture(scope='session', autouse=True)
async def drop_ticket_collection(event_loop):
    yield
    await TicketForQuery.Q.drop_collection(force=True)


def test_query_organization():
    query = (
        Query(name=123) | Query(name__ne=124) & Query(position=1) | Query(position=2)
    )
    data = query.to_query(TicketForQuery)
    value = {
        '$or': [
            {'name': '123'},
            {'$and': [{'name': {'$ne': '124'}}, {'position': 1}]},
            {'position': 2},
        ]
    }
    assert data == value


@pytest.mark.asyncio
async def test_query_result():
    query = [
        TicketForQuery(name='first', position=1),
        TicketForQuery(name='second', position=2),
    ]
    inserted = await TicketForQuery.Q.insert_many(query)
    assert inserted == 2

    query = Query(name='first') | Query(position=1) & Query(name='second')
    data = await TicketForQuery.Q.find_one(query)
    assert data.name == 'first'

    query = Query(position=3) | Query(position=1) & Query(name='second')
    data = await TicketForQuery.Q.find_one(query)
    assert data is None

    query = Query(position=3) | Query(position=2) & Query(name='second')
    data = await TicketForQuery.Q.find_one(query)
    assert data.name == 'second'



import pytest_asyncio
import pytest

from motordantic.models import MongoModel
from motordantic.types import ObjectIdStr
from motordantic.query import Query
from motordantic.aggregation import Sum, Max, Min, Avg, Count
from motordantic.exceptions import MotordanticValidationError


product_types = {1: 'phone', 2: 'book', 3: 'food'}


class Product(MongoModel):
    title: str
    cost: float
    quantity: int
    product_type: str
    config: dict


class ProductImage(MongoModel):
    url: str
    product_id: ObjectIdStr


@pytest_asyncio.fixture(scope='session', autouse=True)
async def innert_producs(event_loop):
    data = [
        Product(
            title=str(i),
            cost=float(i),
            quantity=i,
            product_type=product_types[2] if i != 4 else product_types[1],
            config={'type_id': i},
        )
        for i in range(1, 5)
    ]
    await Product.Q.insert_many(data)
    yield
    await Product.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_aggregation_math_operation():
    max_ = await Product.Q.simple_aggregate(aggregation=Max('cost'))
    assert max_.data == {'cost__max': 4}

    min_ = await Product.Q.simple_aggregate(aggregation=Min('cost'))
    assert min_.data == {'cost__min': 1}

    sum_ = await Product.Q.simple_aggregate(aggregation=Sum('cost'))
    assert sum_.data == {'cost__sum': 10}

    avg_ = await Product.Q.simple_aggregate(aggregation=Avg('cost'))
    assert avg_.data == {'cost__avg': 2.5}

    simple_avg = await Product.Q.aggregate_sum('cost')
    assert simple_avg == 10.0

    simple_max = await Product.Q.aggregate_max('cost')
    assert simple_max == 4

    simple_min = await Product.Q.aggregate_min('cost')
    assert simple_min == 1

    simple_avg = await Product.Q.aggregate_avg('cost')
    assert simple_avg == 2.5


@pytest.mark.asyncio
async def test_aggregation_multiply():
    result_sum = await Product.Q.simple_aggregate(
        aggregation=[Sum('cost'), Sum('quantity')]
    )
    assert result_sum.data == {'cost__sum': 10.0, 'quantity__sum': 10}

    result_max = await Product.Q.simple_aggregate(
        aggregation=[Max('cost'), Max('quantity')]
    )
    assert result_max.data == {'cost__max': 4.0, 'quantity__max': 4}

    result_min = await Product.Q.simple_aggregate(
        aggregation=[Min('cost'), Min('quantity')]
    )
    assert result_min.data == {'cost__min': 1.0, 'quantity__min': 1}

    result_avg = await Product.Q.simple_aggregate(
        aggregation=(Avg('cost'), Avg('quantity'))
    )
    assert result_avg.data == {'cost__avg': 2.5, 'quantity__avg': 2.5}

    result_multiply = await Product.Q.simple_aggregate(
        aggregation=(Avg('cost'), Max('quantity'))
    )
    assert result_multiply.data == {'cost__avg': 2.5, 'quantity__max': 4}

    result_count = await Product.Q.simple_aggregate(aggregation=Count('product_type'))
    assert result_count.data == {'book': {'count': 3}, 'phone': {'count': 1}}

    result_count_agg = await Product.Q.simple_aggregate(
        aggregation=[Count('product_type'), Sum('cost')]
    )
    assert result_count_agg.data == {
        'book': {'cost__sum': 6.0, 'count': 3},
        'phone': {'cost__sum': 4.0, 'count': 1},
    }

    result_sum_and_avg_agg_with_group = await Product.Q.simple_aggregate(
        aggregation=[Avg('cost'), Sum('cost')],
        group_by=['product_type'],
    )
    assert result_sum_and_avg_agg_with_group.data == {
        'phone': {'cost__avg': 4.0, 'cost__sum': 4.0},
        'book': {'cost__avg': 2.0, 'cost__sum': 6.0},
    }

    result_group_by_by_inners = await Product.Q.simple_aggregate(
        group_by=['config.type_id'], aggregation=Count('_id')
    )
    assert result_group_by_by_inners.data == {
        '4': {'count': 1},
        '3': {'count': 1},
        '2': {'count': 1},
        '1': {'count': 1},
    }

    result_sum_and_avg_agg_with_group_many = await Product.Q.simple_aggregate(
        aggregation=[Avg('cost'), Sum('cost')],
        group_by=['product_type', 'quantity'],
    )
    assert result_sum_and_avg_agg_with_group_many.data == {
        'phone|4': {'cost__avg': 4.0, 'cost__sum': 4.0},
        'book|3': {'cost__avg': 3.0, 'cost__sum': 3.0},
        'book|2': {'cost__avg': 2.0, 'cost__sum': 2.0},
        'book|1': {'cost__avg': 1.0, 'cost__sum': 1.0},
    }

    result_agg = await Product.Q.simple_aggregate(
        aggregation=[Avg('cost'), Max('quantity')]
    )
    assert result_agg.data == {'cost__avg': 2.5, 'quantity__max': 4}

    result_not_match_agg = await Product.Q.simple_aggregate(
        Query(title__ne='not_match') & Query(title__startswith='not'),
        aggregation=[Avg('cost'), Max('quantity')],
    )
    assert result_not_match_agg.data == {}


@pytest.mark.asyncio
async def test_raises_invalid_field():
    with pytest.raises(MotordanticValidationError):
        await Product.Q.simple_aggregate(
            title='not_match', aggregation=[Avg('cost123'), Max('quantityzzzz')]
        )


@pytest.mark.asyncio
async def test_raw_aggregate():
    result_raw_group_by_by_inners = await Product.Q.raw_aggregate(
        data=[
            {
                "$group": {
                    "_id": {'type_id': "$config.type_id"},
                    'count': {f'$sum': 1},
                    'names': {'$push': '$title'},
                }
            },
        ],
    )
    assert result_raw_group_by_by_inners == [
        {'_id': {'type_id': 4}, 'count': 1, 'names': ['4']},
        {'_id': {'type_id': 3}, 'count': 1, 'names': ['3']},
        {'_id': {'type_id': 2}, 'count': 1, 'names': ['2']},
        {'_id': {'type_id': 1}, 'count': 1, 'names': ['1']},
    ]



from typing import Optional, TypeVar, Generic, Type, Union, TYPE_CHECKING
from enum import Enum

from bson import ObjectId, DBRef
from bson.errors import InvalidId
from pydantic import parse_obj_as, BaseModel
from pydantic.fields import ModelField
from pydantic.json import ENCODERS_BY_TYPE

__all__ = ('ObjectIdStr', 'Relation', 'RelationInfo', 'RelationTypes')

if TYPE_CHECKING:
    from .models import MongoModel


MONGO_MODEL_TYPE = Type['MongoModel']
T = TypeVar("T")


class ObjectIdStr(str):
    """Field for validate string like ObjectId"""

    type_ = ObjectId
    required = False
    default = None
    validate_always = False
    alias = ''

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> ObjectId:
        if isinstance(v, ObjectId):
            return v
        try:
            return ObjectId(str(v))
        except InvalidId:
            raise ValueError(f"invalid ObjectId - {v}")

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type="string")


class RelationTypes(str, Enum):
    SINGLE = "SINGLE"
    OPTIONAL_SINGLE = "OPTIONAL_SINGLE"
    ARRAY = "ARRAY"


class RelationInfo(BaseModel):
    field: str
    mongo_model_class: MONGO_MODEL_TYPE
    relation_type: RelationTypes


class Relation(Generic[T]):
    def __init__(self, db_ref: DBRef, mongo_model_class: MONGO_MODEL_TYPE):
        self.db_ref = db_ref
        self.mongo_model_class = mongo_model_class

    async def get(self) -> Optional["MongoModel"]:
        result = await self.mongo_model_class.Q.find_one(_id=self.db_ref.id, with_relations_objects=True)  # type: ignore
        return result

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def _validate_for_model(
        cls, v: Union[dict, BaseModel], mongo_model_class: MONGO_MODEL_TYPE
    ) -> 'Relation':
        parsed = (
            mongo_model_class.parse_obj(v)
            if isinstance(v, dict)
            else mongo_model_class.validate(v)
        )
        new_id = (
            parsed._id
            if isinstance(parsed._id, ObjectId)
            else parse_obj_as(
                ObjectIdStr,
                parsed._id,  # type: ignore
            )
        )
        db_ref = DBRef(collection=mongo_model_class.set_collection_name(), id=new_id)
        return cls(db_ref=db_ref, mongo_model_class=mongo_model_class)

    @classmethod
    def validate(cls, v: Union[DBRef, T], field: ModelField):
        mongo_model_class = field.sub_fields[0].type_  # type: ignore
        if isinstance(v, DBRef):
            return cls(db_ref=v, mongo_model_class=mongo_model_class)
        if isinstance(v, Relation):
            return v
        if isinstance(v, dict):
            try:
                return cls(db_ref=DBRef(**v), mongo_model_class=mongo_model_class)
            except TypeError:
                return cls._validate_for_model(v, mongo_model_class)
        if isinstance(v, BaseModel):
            return cls._validate_for_model(v, mongo_model_class)
        raise ValueError(f'invalod type - {v}')

    def to_ref(self):
        return self.db_ref

    def to_dict(self):
        return {"id": str(self.db_ref.id), "collection": self.db_ref.collection}

    @property
    def data(self):
        return self.to_dict()


ENCODERS_BY_TYPE[Relation] = lambda o: o.to_dict()
ENCODERS_BY_TYPE[ObjectIdStr] = lambda o: str(o)



import asyncio
from typing import Type, TYPE_CHECKING, List, Dict

from .types import RelationTypes

__all__ = ('RelationManager',)

if TYPE_CHECKING:
    from .models import MongoModel


MONGO_MODEL_TYPE = Type['MongoModel']
DictStrList = Dict[str, List]


class RelationManager(object):
    """relation manager for get and set data from to to models instances"""

    __slots__ = ('mongo_model_class', 'relation_fields')

    def __init__(self, mongo_model_class: MONGO_MODEL_TYPE):
        self.mongo_model_class = mongo_model_class
        self.relation_fields = self._get_relation_fields(mongo_model_class)

    @classmethod
    def _get_relation_fields(cls, mongo_model_class: MONGO_MODEL_TYPE) -> dict:
        return mongo_model_class.__db_refs__ or {}

    def _relation_data_setter(
        self, mongo_model: 'MongoModel', data: dict
    ) -> 'MongoModel':
        """data setter

        Args:
            mongo_model (MongoModel): mongo model insatance
            data (dict): relations objects map

        Returns:
            MongoModel: updated mongo model
        """
        for field, relation_info in self.relation_fields.items():
            relation_attr = getattr(mongo_model, field)
            if not relation_attr:
                continue
            if relation_info.relation_type == RelationTypes.ARRAY:
                relation_value = []
                for rel in relation_attr:
                    v = data[field].get(str(rel.db_ref.id))
                    if v:
                        relation_value.append(v)
            else:
                relation_value = data[field].get(str(relation_attr.db_ref.id))
            setattr(mongo_model, field, relation_value)
        return mongo_model

    async def get_relation_objects(self, pre_relation: DictStrList) -> dict:
        futures = []
        field_map = {}
        for index_, row in enumerate(pre_relation.items(), start=1):
            field, ids = row
            mongo_model_class = self.relation_fields[field].mongo_model_class
            futures.append(
                asyncio.ensure_future(
                    mongo_model_class.Q.find(_id__in=ids, with_relations_objects=True)
                )
            )
            field_map[index_] = field

        relation_objects = {}
        for index_, field_relation_result in enumerate(
            asyncio.as_completed(futures), start=1
        ):
            field = field_map[index_]
            relation_objects[field] = {
                str(o._id): o for o in await field_relation_result
            }
        return relation_objects

    def _get_pre_relation(
        self, mongo_model_instances: List['MongoModel']
    ) -> DictStrList:
        pre_relation: DictStrList = {field: [] for field in self.relation_fields}
        for mongo_model_instance in mongo_model_instances:
            for field in self.relation_fields:
                attr = getattr(mongo_model_instance, field)
                if isinstance(attr, list):
                    ids = tuple(row.db_ref.id for row in attr)
                else:
                    ids = (attr.db_ref.id,) if attr else tuple()
                if ids:
                    pre_relation[field].extend(ids)
        return {f: list(set(items)) for f, items in pre_relation.items()}

    async def map_relation_for_single(
        self, mongo_model_instance: 'MongoModel'
    ) -> 'MongoModel':
        """map relation data to mongo model instanc

        Args:
            mongo_model_instances (List[MongoModel]): list of instances from MongoModel

        Returns:
            MongoModel: mapped mongo model
        """
        pre_relation: DictStrList = self._get_pre_relation([mongo_model_instance])
        relation_objects = await self.get_relation_objects(pre_relation)
        return self._relation_data_setter(mongo_model_instance, relation_objects)

    async def map_relation_for_array(self, result: List) -> List:
        """map relations data for _find method list result

        Args:
            result (List): _find query result converted to list

        Returns:
            List: mapped list
        """
        pre_relation: DictStrList = self._get_pre_relation(result)
        relation_objects = await self.get_relation_objects(pre_relation)
        generated_result = [
            self._relation_data_setter(r, relation_objects) for r in result
        ]
        return generated_result



from io import open
from setuptools import setup, find_packages


def read(f):
    return open(f, "r").read()


setup(
    name="motordantic",
    version='0.0.1a2',
    packages=find_packages(exclude=("tests", "docs", "examples")),
    install_requires=[
        "pydantic>=1.9,<2",
        "pymongo>=3.10.1<=4.0.1",
        "motor==3.0.0",
    ],
    description="Mongo ODM, based on motor+pydantic",
    author="bogdanjz",
    author_email="bzdv.dn@gmail.com",
    url="https://github.com/bogdanjz/motordantic",
    license="MIT",
    python_requires=">=3.8",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
)



import copy
from json import dumps
from typing import Generator, List, Union, Any, Tuple, List, TYPE_CHECKING, Type


__all__ = (
    "Query",
    "QueryCombination",
    "FindResult",
    "SimpleAggregateResult",
)


if TYPE_CHECKING:
    from .models import MongoModel

MONGO_MODEL_TYPE = Type['MongoModel']


def _validate_query_data(model: MONGO_MODEL_TYPE, query: dict) -> dict:
    return model._validate_query_data(query)


class QueryNodeVisitor(object):
    """Base visitor class for visiting Query-object nodes in a query tree."""

    def prepare_combination(
        self, combination: 'QueryCombination'
    ) -> Union['QueryCombination', dict]:
        """Called by QueryCombination objects."""
        return combination

    def visit_query(self, query: 'Query') -> Union['Query', dict]:
        """Called by (New)Query objects."""
        return query


class SimplificationVisitor(QueryNodeVisitor):
    def __init__(self, model: MONGO_MODEL_TYPE):
        self.model = model

    def prepare_combination(
        self, combination: 'QueryCombination'
    ) -> Union['QueryCombination', dict]:
        if combination.operation == combination.AND:
            # The simplification only applies to 'simple' queries
            if all(isinstance(node, Query) for node in combination.children):
                queries = [n.query for n in combination.children]
                query = self._query_conjunction(queries)
                return {"$and": query}

        return combination

    def _query_conjunction(self, queries):
        """Merges query dicts - effectively &ing them together."""
        combined_query = []
        for query in queries:
            query = _validate_query_data(self.model, query)
            combined_query.append(copy.deepcopy(query))
        return combined_query


class QueryCompilerVisitor(QueryNodeVisitor):
    """Compiles the nodes in a query tree to a PyMongo-compatible query
    dictionary.
    """

    def __init__(self, model: MONGO_MODEL_TYPE):
        self.model = model

    def prepare_combination(
        self, combination: 'QueryCombination'
    ) -> Union['QueryCombination', dict]:
        operator = "$and"
        if combination.operation == combination.OR:
            operator = "$or"
        return {operator: combination.children}

    def visit_query(self, query: 'Query') -> Union['Query', dict]:
        data = _validate_query_data(self.model, query.query)
        return data


class QueryNode(object):
    """Base class for nodes in query trees."""

    AND = 0
    OR = 1

    def to_query(self, model: MONGO_MODEL_TYPE) -> dict:
        query = self.accept(SimplificationVisitor(model))
        if not isinstance(query, dict):
            query = query.accept(QueryCompilerVisitor(model))
        return query

    def accept(self, visitor):
        raise NotImplementedError

    def _combine(self, other, operation):
        """Combine this node with another node into a QueryCombination
        object.
        """
        # If the other Query() is empty, ignore it and just use `self`.
        if getattr(other, "empty", True):
            return self

        # Or if this Q is empty, ignore it and just use `other`.
        if self.empty:
            return other

        return QueryCombination(operation, [self, other])

    @property
    def empty(self):
        return False

    def __or__(self, other):
        return self._combine(other, self.OR)

    def __and__(self, other):
        return self._combine(other, self.AND)


class QueryCombination(QueryNode):
    def __init__(self, operation, children):
        self.operation = operation
        self.children = []
        for node in children:
            # If the child is a combination of the same type, we can merge its
            # children directly into this combinations children
            if isinstance(node, QueryCombination) and node.operation == operation:
                self.children += node.children
            else:
                self.children.append(node)

    def __repr__(self):
        op = " & " if self.operation is self.AND else " | "
        return "(%s)" % op.join([repr(node) for node in self.children])

    def __bool__(self):
        return bool(self.children)

    def accept(self, visitor) -> Union['QueryCombination', dict]:
        for i in range(len(self.children)):
            if isinstance(self.children[i], QueryNode):
                self.children[i] = self.children[i].accept(visitor)

        return visitor.prepare_combination(self)

    @property
    def empty(self):
        return not bool(self.children)

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.operation == other.operation
            and self.children == other.children
        )


class Query(QueryNode):
    """A simple query object, used in a query tree to build up more complex
    query structures.
    """

    def __init__(self, **query):
        self.query = query

    def __repr__(self):
        return "Query(**%s)" % repr(self.query)

    def __bool__(self):
        return bool(self.query)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.query == other.query

    def accept(self, visit: 'QueryNodeVisitor') -> Union['Query', dict]:
        return visit.visit_query(self)

    @property
    def empty(self) -> bool:
        return not bool(self.query)


class FindResult(object):
    __slots__ = ('_data', 'mongo_model_class')

    def __init__(
        self,
        mongo_model_class: MONGO_MODEL_TYPE,
        data: list,
    ):
        self._data = data
        self.mongo_model_class = mongo_model_class

    # @handle_and_convert_connection_errors
    def __iter__(self):
        for obj in self._data:
            yield obj

    def __next__(self):
        return next(self.__iter__())

    @property
    def data(self) -> List:
        return [obj.data for obj in self.__iter__()]

    @property
    def generator(self) -> Generator:
        return self.__iter__()

    @property
    def data_generator(self) -> Generator:
        for obj in self.__iter__():
            yield obj.data

    @property
    def list(self) -> List:
        return list(self.__iter__())

    def json(self) -> str:
        return dumps(self.data)

    def first(self) -> Any:
        return next(self.__iter__())

    def serialize(
        self, fields: Union[Tuple, List], to_list: bool = True
    ) -> Union[Tuple, List]:
        return (
            [obj.serialize(fields) for obj in self.__iter__()]
            if to_list
            else tuple(obj.serialize(fields) for obj in self.__iter__())
        )

    def serialize_generator(self, fields: Union[Tuple, List]) -> Generator:
        for obj in self.__iter__():
            yield obj.serialize(fields)

    def serialize_json(self, fields: Union[Tuple, List]) -> str:
        return dumps(self.serialize(fields))


class SimpleAggregateResult(object):
    __slots__ = ('_data', 'mongo_model_class')

    def __init__(
        self,
        mongo_model_class: MONGO_MODEL_TYPE,
        data: dict,
    ):
        self._data = data
        self.mongo_model_class = mongo_model_class

    def json(self) -> str:
        return dumps(self._data)

    @property
    def data(self) -> dict:
        return self._data



import os
from typing import Optional

from motor.motor_asyncio import AsyncIOMotorClient

from .singleton import Singleton

DEFAULT_ENV_NAME = 'default'


_connections: dict = {}


class DBConnection(object, metaclass=Singleton):
    __slots__ = (
        'address',
        'db_name',
        'max_pool_size',
    )

    _connections: dict = {}

    def __init__(
        self,
        address: str,
        db_name: str,
        max_pool_size: int = 250,
        env_name: str = DEFAULT_ENV_NAME,
    ):
        self.address = address
        self.db_name = db_name
        self.max_pool_size = max_pool_size
        if env_name not in _connections:
            _connections[env_name] = self

    def _init_mongo_connection(self, connect: bool = False) -> AsyncIOMotorClient:
        client = AsyncIOMotorClient(self.address, connect=connect)
        return client

    def _get_motor_client(self) -> AsyncIOMotorClient:
        if os.getpid() in self._connections:
            return self._connections[os.getpid()]
        else:
            mongo_connection = self._init_mongo_connection()
            self._connections[os.getpid()] = mongo_connection
            return mongo_connection


def connect(
    address: str,
    database_name: str,
    ssl: bool = False,
    max_pool_size: int = 100,
    ssl_cert_path: Optional[str] = None,
    server_selection_timeout_ms: int = 60000,
    connect_timeout_ms: int = 30000,
    socket_timeout_ms: int = 60000,
    env_name: Optional[str] = DEFAULT_ENV_NAME,
) -> None:
    """init connection to mongodb

    Args:
        address (str): full connection string
        dbname (str): mongo db name
        ssl (bool, optional): flag for ssl cert. Defaults to False.
        max_pool_size (int, optional): max connection pool. Defaults to 100.
        ssl_cert_path (Optional[str], optional): path to ssl cert. Defaults to None.
        server_selection_timeout_ms (int, optional): ServerSelectionTimeoutMS. Defaults to 60000.
        connect_timeout_ms (int, optional): ConnectionTimeoutMS. Defaults to 30000.
        socket_timeout_ms (int, optional): SocketTimeoutMS. Defaults to 60000.
        env_name (Optional[str], optional): connection env name. Defaults to None.
    """
    os.environ['MOTORDANTIC_DATABASE'] = database_name
    os.environ['MOTORDANTIC_ADDRESS'] = address
    os.environ['MOTORDANTIC_ENV_NAME'] = env_name or DEFAULT_ENV_NAME
    _ = DBConnection(
        address=address,
        db_name=database_name,
        env_name=env_name or DEFAULT_ENV_NAME,
    )
